@@ -5,7 +5,8 @@ from __future__ import absolute_import
 from future.utils import python_2_unicode_compatible
 
 import pykeepass.entry
-from collections import namedtuple
+
+from pykeepass.exceptions import BinaryError
 
 # FIXME python2
 @python_2_unicode_compatible
@@ -19,6 +20,7 @@ class Attachment(object):
 
     @property
     def id(self):
+        """str: get or set id of binary the attachment points to"""
         return int(self._element.find('Value').attrib['Ref'])
 
     @id.setter
@@ -27,6 +29,7 @@ class Attachment(object):
 
     @property
     def filename(self):
+        """str: get or set filename attachment"""
         return self._element.find('Key').text
 
     @filename.setter
@@ -35,15 +38,20 @@ class Attachment(object):
 
     @property
     def entry(self):
+        """Entry: get entry this attachment is associated with"""
         ancestor = self._element.getparent()
         return pykeepass.entry.Entry(element=ancestor, kp=self._kp)
 
     @property
-    def data(self):
+    def binary(self):
+        """bytes: get binary this attachment points to"""
         try:
             return self._kp.binaries[self.id]
         except IndexError:
-            raise AttachmentError('No such attachment id')
+            raise BinaryError('No such binary with id {}'.format(self.id))
+
+    data = binary
 
     def delete(self):
+        """delete this attachment"""
         self._element.getparent().remove(self._element)
